/* Functions to help parse a C declaration. */

/* 
   Copyright (C) 1998  Ben K. Bullock

   Cfunctions is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   Cfunctions is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cfunctions; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "wt.h"
#include "sys_or_exit.h"
#include "error_msg.h"
#include "argument.h"

#ifdef HEADER

#include "tags.h"

/* The type of an argument. */

struct type
{
  char * name;
  struct type * next, * prev;
  unsigned line;              /* Line number it was seen on. */
};

/* Some argument types are shared between several arguments.  For
   example in `int i, j;', the type `int' is shared between i and j.
   This structure stores those relationships. */

struct shared_type
{
  struct type * t;
  int ref_count;
};

/* An argument. */

struct arg
{
  enum {TYPE, POINTER, SUFFIX} parse_state; /* For clarity. */
  struct type * name;         /* Name of this argument. */
  struct shared_type * types; /* Type. */
  struct type * pointers;     /* Pointer prefixes. */
  struct type * suffixes;     /* Array suffixes. */
  unsigned line;              /* Line number of the name. */
  struct arg * prev, * next;  /* For shared arguments. */
};

#endif /* HEADER */

static const struct type type_null;
static const struct arg  arg_null;

/* Create a new arg structure. */

struct arg * 
arg_start (void)
{
  struct arg * a;

  a = calloc_or_exit (1, sizeof (struct arg));
  a->types = calloc_or_exit (1, sizeof (struct shared_type));
  a->types->ref_count++;
  if ( a->next || a->prev )
    bug (HERE, "fouled up again");
  return a;
}

/* Start a new arg with the same type as another one. */

struct arg * 
arg_share (struct arg * a)
{
  struct arg * b;

  b = calloc_or_exit (1, sizeof (struct arg));
  b->types = a->types;
  b->types->ref_count++;
  b->parse_state = POINTER;
  b->prev = a;
  if ( b->next )
    bug (HERE, "fouled up again");
  a->next = b;
  return b;
}

/* Add to a linked list of types. */

static void
type_list_add (struct type ** a, struct type * b)
{
  if (* a)
    {
      (* a)->next = b;
      b->prev = * a;
    }
  * a = b;
}

/* Add another string to an argument. */

void
arg_add (struct arg * a, const char * t, unsigned line)
{
  unsigned t_len;
  struct type * x;
  t_len = strlen (t);
#if 0
  x = calloc_or_exit (1, sizeof (struct type) + t_len + 1);
  x->name = (char *) x + sizeof (struct type);
#else
  x = calloc_or_exit (1, sizeof (struct type));
  x->name = malloc_or_exit (t_len + 1);
#endif
  strcpy (x->name, t);
  x->line = line;
  if ( t[0] == '*')
    a->parse_state = POINTER;
  switch (a->parse_state)
    {
    case SUFFIX:
#if 0
      if (! strchr (t, '['))
        bug (HERE, "no [ in suffix");
#endif
      type_list_add (& a->suffixes, x);
      break;

    case POINTER:
      type_list_add (& a->pointers, x);
      break;

    case TYPE:
      if (a->types->ref_count > 1)
        bug (HERE, "addition to a shared type list");
      else
        type_list_add (& a->types->t, x);
      break;
    default:
      bug (HERE, "bad value in switch");
    }
}

/* Move a word out of a type list and into the name of an argument. */

static void
type_list_move (struct arg * a, struct type ** t)
{
  a->name = * t;
  if ((* t)->prev)
    (* t)->prev->next = NULL;
  * t = (* t)->prev;
}

/* For a particular argument, make the last-seen word the name
   of the argument. */

void
arg_put_name (struct arg * a)
{
  if (a->name)
    return;
  switch (a->parse_state)
    {
    case POINTER:
      if (! a->pointers)
        bug (HERE, "attempt to move non-existent name from pointer list");
      type_list_move (a, & a->pointers);
      break;
    case TYPE:
      if (! a->types->t)
        return;
      if (a->types->ref_count > 1)
        bug (HERE, "attempt to move a shared type to arg name");
      type_list_move (a, & a->types->t);
      break;
    case SUFFIX:
      bug (HERE, "attempt to move a suffix to arg name");
    }

  /* Anything which comes now must be a suffix. */

  a->parse_state = SUFFIX;
}

/* Free a list of types.  Things are arranged so that the string which
   is part of the type is always at the end of the type, so it is not
   necessary to free that string separately. */

static void
type_free (struct type * t)
{
  if (! t)
    return;
  while (t->prev)
    t = t->prev;
#if 0
  while ((t = t->next))
    free (t->prev);
#else
  while (t->next)
    {
#if 0
      printf ("Freeing %s\n", t->name);
#endif
      free (t->name);
      t = t->next;
      free (t->prev);
    }
#if 0
  printf ("Freeing %s\n", t->name);
#endif
  free (t->name);
  free (t);
#endif
}

/* Free the memory associated with an argument. */

void
arg_free (struct arg * a)
{
  a->types->ref_count--;
  if (a->types->ref_count == 0)
    {
#if 0
      printf ("zero refs\n");
#endif
      type_free (a->types->t);
      free (a->types);
    }
  if (a->name)
    {
#if 0
      printf ("Freeing `%s'\n", a->name->name);
#endif
      free (a->name->name);
      free (a->name);
    }
  type_free (a->pointers);
  type_free (a->suffixes);
  if (a->next)
    a->next->prev = NULL;
  if (a->prev)
    arg_free (a->prev);
  free (a);
}

extern void cpp_eject (unsigned);

/* Print a list of array suffixes to the specified stream. */

static void
suffix_fprint (FILE * f, struct type * t)
{
  if (! t)
    return;
  while (t->prev)
    t = t->prev;
  fprintf (f, "[]");
  t = t->next;
  while (t)
    {
      fprintf (f, "%s", t->name);
      t = t->next;
    }
}


/* Print a list of types to the specified stream. */

static void
type_fprint (FILE * f, struct type * t, int do_extern)
{
  int did_extern = 0;
  if (! t)
    return;
  while (t->prev)
    t = t->prev;
  do 
    {
      if (t->name[0] == '@')
        {
          unsigned cpp_id;
          if (strncmp (t->name + 1, "CPP", 3) == 0)
            {
              cpp_id = atoi (t->name + 4);

              /* I removed the `extern' printing bit. */

              cpp_eject (cpp_id);
            }
          else
            bug (HERE, "unknown @ escape `%s'", t->name);
        }
      else
        {
          if (do_extern && ! did_extern)
            {
              fprintf (f, "extern ");
              did_extern = 1;
            }
          fprintf (f, "%s ", t->name);
        }
      t = t->next;
    }
  while (t);
}

/* Print just one argument. */

void
arg_fprint (FILE * f, struct arg * a)
{
  if (a->types->t)
    type_fprint (f, a->types->t, 0);
  else
    {
      if (warns.implicit_int)
        line_warning ("function `%s' is implicit int", a->name->name);
      fprintf (outfile, "int /* default */");
    }
  type_fprint (f, a->pointers, 0);
  if (a->name)
    fprintf (f, "%s", a->name->name);
  suffix_fprint (f, a->suffixes);
}

/* Print all the arguments in a list of them.  This is used for
   printing a list of globals, so the `extern' prefix is always used
   here. */

void
arg_fprint_all (FILE * f, struct arg * a, int do_extern)
{
  if (a->types->t)
    type_fprint (f, a->types->t, do_extern);
  while (a->prev)
    a = a->prev;
  for (; a; a = a->next)
    {
      type_fprint (f, a->pointers, 0);
      if (a->name)
        fprintf (f, "%s", a->name->name);
      suffix_fprint (f, a->suffixes);
      if (a->next)
        fprintf (f, ", ");
    }
}

/* Make all the tags for a particular argument. */

void
arg_tag (struct arg * a, enum tag_type t)
{
  while (a->prev)
    a = a->prev;
  for (; a; a = a->next)
    if (a->name)
      tag_make (a->name->name, t, a->name->line);
}

/* If there is anything tagable in the type field of `a' then make it
   a tag.  This is for struct, union and enum name tagging only. */

void
arg_tagable (struct arg * a)
{
  struct type * t = a->types->t;
  if (! t)
    return;
  if (strcmp (t->name, "union") == 0)
    return;
  if (strcmp (t->name, "struct") == 0)
    return;
  if (strcmp (t->name, "enum") == 0)
    return;
  if (strcmp (t->name, "typedef") == 0)
    return;
  tag_make (t->name, TAG_STRUCT_UNION_ENUM, t->line);
}
